/*
 * OHCI HCD (Host Controller Driver) for USB.
 *
 * Bus Glue for EP93XX USB Controller
 *
 * Based on ohci-lh7a404 driver and previous EP93XX driver.
 *
 *
 * This file is licensed under the GPL.
 */

#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/arch/hardware.h>

extern int usb_disabled(void);

/*-------------------------------------------------------------------------*/

static int ep93xx_start_hc(struct ohci_hcd *ohci, struct platform_device *pdev)
{
#if defined(CONFIG_ARCH_EDB9315)
	unsigned int uiPBDR, uiPBDDR;
#endif

	pr_debug("Starting EP93XX OHCI USB Controller\n");
     
	/*
	 * Configure the power sense and control lines.  Place the USB
	 * host controller in reset.
	 */
#if defined(CONFIG_ARCH_EDB9315)
	/*
	 * For EDB9315 boards, turn on USB by clearing EGPIO9 (Port B, Bit 1).
	 */
	uiPBDR = inl(GPIO_PBDR) & 0xfd;
	outl(uiPBDR, GPIO_PBDR );

	uiPBDDR = inl(GPIO_PBDDR) | 0x02;
	outl(uiPBDDR, GPIO_PBDDR );
#endif

	/*
	 * Now, carefully enable the USB clock, and take
	 * the USB host controller out of reset.
	 */
	writel(readl((void *)SYSCON_PWRCNT) | SYSCON_PWRCNT_USHEN,(void *)SYSCON_PWRCNT);

        return ohci_init(ohci);
}

static void ep93xx_stop_hc(struct platform_device *pdev)
{
#if defined(CONFIG_ARCH_EDB9315)
	unsigned int uiPBDR;
#endif

	pr_debug("Stopping EP93XX OHCI USB Controller\n");

	/* Shut down the USB block */
	writel(readl((void *)SYSCON_PWRCNT) & (~SYSCON_PWRCNT_USHEN),(void *)SYSCON_PWRCNT);

#if defined(CONFIG_ARCH_EDB9315)
	/*
	 * For EDB9315 boards, turn off USB by setting EGPIO9 (Port B, Bit 1).
	 */
	uiPBDR = inl(GPIO_PBDR) | 0x02;
	outl(uiPBDR, GPIO_PBDR );
#endif 
}

/*-------------------------------------------------------------------------*/

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */


/**
 * usb_hcd_ep93xx_probe - initialize EP93XX HCD
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 */
static int usb_hcd_ep93xx_probe (const struct hc_driver *driver,
				 struct platform_device *pdev)
{
	int retval = 0;
	struct usb_hcd *hcd = 0;
	struct ohci_hcd *ohci;
	struct resource *res;

	hcd = usb_create_hcd(driver, &pdev->dev, pdev->dev.bus_id);

	if (!hcd) {
		retval = -ENOMEM;
		goto err0;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	if (!res) {
		retval = -ENODEV;
		goto err1;
	}

	hcd->rsrc_start = HwRegToVirt(res->start);
	hcd->rsrc_len = res->end - res->start + 1;

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed");
		retval = -EBUSY;
		goto err1;
	}

	hcd->regs = (void __iomem *)hcd->rsrc_start;

	ohci = hcd_to_ohci(hcd);
	ohci_hcd_init(ohci);
	retval = ep93xx_start_hc(ohci, pdev);
	if (retval < 0)
		goto err2;

	retval = usb_add_hcd(hcd, platform_get_irq(pdev, 0), SA_INTERRUPT);
	if (retval == 0)
		return retval;

	ep93xx_stop_hc(pdev);
err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
err1:
	usb_put_hcd(hcd);
err0:
	return retval;
}

/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_ep93xx_remove - shutdown processing for EP93XX HCD
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_ep93xx_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
static void usb_hcd_ep93xx_remove (struct usb_hcd *hcd, struct platform_device *pdev)
{
	usb_remove_hcd(hcd);
	ep93xx_stop_hc(pdev);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
}

/*-------------------------------------------------------------------------*/

static int __devinit
ohci_ep93xx_start (struct usb_hcd *hcd)
{
	struct ohci_hcd	*ohci = hcd_to_ohci (hcd);
	int		ret;

	ohci_dbg(ohci, "ohci_ep93xx_start, ohci:%p", ohci);

	if ((ret = ohci_run(ohci)) < 0) {
		err("can't start %s", hcd->self.bus_name);
		ohci_stop(hcd);
		return ret;
	}

	return 0;
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ohci_ep93xx_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"EP93xx OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			ohci_irq,
	.flags =		HCD_USB2,

	/*
	 * basic lifecycle operations
	 */
	.start =		ohci_ep93xx_start,
#ifdef	CONFIG_PM
	/* suspend:		ohci_ep93xx_suspend,  -- tbd */
	/* resume:		ohci_ep93xx_resume,   -- tbd */
#endif /*CONFIG_PM*/
	.stop =			ohci_stop,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
};

/*-------------------------------------------------------------------------*/

static int ohci_hcd_ep93xx_drv_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);

	pr_debug ("In ohci_hcd_ep93xx_drv_probe");

	return usb_hcd_ep93xx_probe(&ohci_ep93xx_hc_driver, pdev);
}

static int ohci_hcd_ep93xx_drv_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	usb_hcd_ep93xx_remove(hcd, pdev);
	dev_set_drvdata(dev, NULL);
	return 0;
}

	/*TBD*/
/*
static int ohci_hcd_ep93xx_drv_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	return 0;
}
static int ohci_hcd_ep93xx_drv_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	return 0;
}
*/

static struct device_driver ohci_hcd_ep93xx_driver = {
	.name           = "ep93xxusb",
	.bus            = &platform_bus_type,
	.probe          = ohci_hcd_ep93xx_drv_probe,
	.remove         = ohci_hcd_ep93xx_drv_remove,
	/*.suspend      = ohci_hcd_ep93xx_drv_suspend, */
	/*.resume       = ohci_hcd_ep93xx_drv_resume, */
};

static int __init ep93xxohci_init(void)
{
	int retval;

	pr_debug(DRIVER_INFO " (EP93XX_USB)");

	if (usb_disabled())
		return -ENODEV;

	pr_debug("block sizes: ed %d td %d\n",
		 sizeof(struct ed), sizeof(struct td));

	retval = driver_register(&ohci_hcd_ep93xx_driver);

	return retval;
}

static void __exit ep93xxohci_cleanup(void)
{
	driver_unregister(&ohci_hcd_ep93xx_driver);
}

module_init(ep93xxohci_init);
module_exit(ep93xxohci_cleanup);
